/******************************************************************************
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        iqs7221e-example-code.ino                                     *
 * @brief       IQS7221E Hall Rotation Example code                           *
 * @details     This example demonstrates how to write the desired settings   *
 *              to the IQS7221E to use the Hall-Rotation EV-Kit, and how      *
 *              to read the encoder interval output and touch events.         *
 *                                                                            *
 *              Data is displayed over serial communication with the          *
 *              following outputs:                                            *
 *                  - Hall Scroll with I2C interval output                    *
 *                    (DEMO_I2C_OR_ENCODER = 0)                               *
 *                  - Hall Scroll by reading Encoder output pins              *
 *                    (DEMO_I2C_OR_ENCODER = 1)                               *
 *                  - Touch events                                            *
 *                  - Power Mode Switching (if enabled)                       *
 *                  - Force Communication when 'f' is sent over Serial        *
 *              When HID is enabled (DEMO_HID_ON 1), mouse scroll outputs     *
 *              will be streamed over HID (Keep Serial Monitor open)          *
 *                                                                            *
 * @author      Azoteq PTY Ltd                                                *
 * @version     v1.2                                                          *
 * @date        2023-07-14                                                    *
 ******************************************************************************/
#include <Arduino.h>
#include "src\IQS7221E\IQS7221E.h"

/*** Defines ***/
#define DEMO_IQS7221E_ADDR          0x56
#define DEMO_IQS7221E_POWER_PIN     4
#define DEMO_IQS7221E_RDY_PIN       7

/* Encoder Input Pins (needs to be interrupt-capable pins) */
#define DEMO_ENC_PIN_A              0
#define DEMO_ENC_PIN_B              1

/* UI Selection */
#define DEMO_I2C_OR_ENCODER         0 // 0 - I2C, 1 - Encoder
#define DEMO_HID_ON                 1 // Turn on HID and give mouse scroll commands over USB.

#if DEMO_I2C_OR_ENCODER != 0
    #include "Encoder.h"
#endif

#if DEMO_HID_ON != 0
    #include <Mouse.h>
#endif

/*** Instances ***/
IQS7221E iqs7221e; // Create an IQS7221E instance
#if DEMO_I2C_OR_ENCODER != 0
    Encoder demo_encoder(DEMO_ENC_PIN_A, DEMO_ENC_PIN_B);
#endif

/*** Global Variables ***/
bool show_data = false;
int32_t enc_interval = 0;
// Get the number of intervals from the IQS7221E_init settings file
const uint16_t max_intervals = NUMBER_INTERVALS_0 + (NUMBER_INTERVALS_1 << 8);
uint8_t running_display_counter = 0;

void setup()
{
    /* Start Serial Communication */
    Serial.begin(115200);
    while(!Serial);
    Serial.println("Start Serial communication");
    delay(200);

    /* Power On IQS7221E */
    pinMode(DEMO_IQS7221E_POWER_PIN, OUTPUT);
    delay(200);
    digitalWrite(DEMO_IQS7221E_POWER_PIN, LOW);
    delay(200);
    digitalWrite(DEMO_IQS7221E_POWER_PIN, HIGH);

    /* Initialize the IQS7221E with input parameters device address and RDY pin */
    iqs7221e.begin(DEMO_IQS7221E_ADDR, DEMO_IQS7221E_RDY_PIN);
    Serial.println("IQS7221E Ready");
    delay(1);

    /* Start Mouse control */
    #if DEMO_HID_ON != 0
        Mouse.begin();
    #endif
}

void loop()
{
    iqs7221e.run(); // Runs the IQS7221E

    force_comms_and_reset(); // function to initialize a force communication window.

    // Check for interval changes on the quadrature pins
    #if DEMO_I2C_OR_ENCODER == 1
        check_encoder_pins();
    #endif

    /* Process data read from IQS7221E when new data is available (RDY Line Low) */
    if (iqs7221e.new_data_available)
    {
        if (show_data)
        {
            show_iqs7221e_data();
        }
        else
        {
            print_Heading();
            #if DEMO_I2C_OR_ENCODER == 0
                check_current_interval();
            #endif
            check_button_state();

            check_power_mode();
        }

        iqs7221e.new_data_available = false;
    }
}

/* Function to check when the Current power mode of the IQS7221E changed */
void check_power_mode(void)
{
    // read current power mode
    switch (iqs7221e.get_PowerMode())
    {
        case IQS7221E_HIGH_ACCURACY:
            Serial.println("HIGH ACCURACY");
            break;
        case IQS7221E_NORMAL_POWER:
            Serial.println("NORMAL POWER");
            break;
        case IQS7221E_LOW_POWER:
            Serial.println("LOW POWER");
            break;
        case IQS7221E_ULP:
            Serial.println("ULP");
            break;
        case IQS7221E_HALT_MODE:
            Serial.println("HALT MODE");
            break;
        default:
            break;
    }
}

void print_Heading(void)
{
    /* Check if it is necessary to display the heading */
    if (running_display_counter == 11 || running_display_counter == 0)
    {
        /* If the I2C feedback from the encoder is enabled, print the extra
        heading for the serial feedback */
        #if DEMO_I2C_OR_ENCODER == 0
            Serial.println("\nENC Interval:\tButton:\t\tPower Mode:");
        #else
            Serial.println("\nButton:\t\tPower Mode:");
        #endif

        running_display_counter = 1;
    }

    running_display_counter++;
}

/* Function to check the Proximity and Touch states of the self-capacitance button */
void check_button_state(void)
{
    /* check if the touch state bit is set for channel 1 */
    if (iqs7221e.button_touchState())
    {
        Serial.print("Touch\t\t");
    }
    /* check if the proximity state bit is set for channel 1 */
    else if (iqs7221e.button_proxState())
    {
        Serial.print("Prox\t\t");
    }
    /* None state if neither the touch nor proximity bit was set to 1 */
    else
    {
        Serial.print("None\t\t");
    }
}

/* Check the current interval of the Quadrature */
void check_current_interval(void)
{
    if (iqs7221e.quad_interval_changed())
    {
        #if DEMO_HID_ON != 0
        if (iqs7221e.quad_interval_direction())
        {
            Mouse.move(0, 0, 1); // scroll mouse up by 1.
        }
        else
        {
            Mouse.move(0, 0, -1); // scroll mouse down by 1.
        }
        #endif
    }

    Serial.print(iqs7221e.quad_interval_value());
    Serial.print("\t\t");

}

/* Read the encoder pins and determine an up/down interval */
#if DEMO_I2C_OR_ENCODER != 0
void check_encoder_pins(void)
{
    // Read the encoder lines for quadrature pin changes
    // Note: both quadrature pins change on a single interval change.
    // `readAndReset()` will return +/-2 for a single interval change.
    int32_t enc_change = demo_encoder.readAndReset();

    /* increment the encoder interval counter */
    enc_interval += enc_change;

    /* Roll over the interval */
    if (enc_interval >= (2 * max_intervals))
    {
        enc_interval = 0;
    }
    else if (enc_interval < 0)
    {
        enc_interval = (2 * max_intervals) - 1;
    }

    if (enc_change != 0)
    {
        #if DEMO_HID_ON != 0
        if ((enc_change > 0) && (enc_interval % 2 == 0))
        {
            Mouse.move(0, 0, 1); // scroll mouse up by 1.
        }
        else if ((enc_change < 0) && (enc_interval % 2 == 0))
        {
            Mouse.move(0, 0, -1); // scroll mouse down by 1.
        }
        #endif

        Serial.print("ENC INTERVAL:");
        Serial.println(enc_interval);
    }
}
#endif

/* Force the IQS7221E to open a RDY window and read the current state of the
device or request a software reset */
void force_comms_and_reset(void)
{
  char message = read_message();

  /* If an 'f' was received over serial, open a forced communication window and
  print the new data received */
  if(message == 'f')
  {
    iqs7221e.force_I2C_communication(); // prompt the IQS7221E
    show_data = true;
  }

  /* If an 'r' was received over serial, request a software(SW) reset */
  if(message == 'r')
  {
    Serial.println("Software Reset Requested!");
    iqs7221e.force_I2C_communication(); // Request a RDY window
    iqs7221e.iqs7221e_state.state = IQS7221E_STATE_SW_RESET;
    running_display_counter = 0;
  }
}

/* Read message sent over serial communication */
char read_message(void)
{
  while (Serial.available())
  {
    if(Serial.available() > 0)
    {
      return  Serial.read();
    }
  }

  return '\n';
}

/* shows the current IQS7221E data */
void show_iqs7221e_data()
{
    if (show_data)
    {
        Serial.println("\n********************************************");
        Serial.println("******* IQS7221E DATA - FORCED COMMS *******");
        Serial.print("********************************************");
        running_display_counter = 0;
        print_Heading();
        #if DEMO_I2C_OR_ENCODER != 0
        Serial.print("ENC INTERVAL:");
        Serial.println(enc_interval);
        #else
        check_current_interval();
        #endif
        check_button_state();
        check_power_mode();

        Serial.println("********************************************");
        show_data = false;
    }
}
