/******************************************************************************
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        IQS7221E.cpp                                                  *
 * @brief       This file contains the constructors and methods which allow   *
 *              ease of use of an IQS7221E Hall-rotation sensor. The IQS7221E *
 *              is a capacitive touch and Hall-rotation Integrated Circuit    *
 *              (IC) which provides multiple-channel functionality. This      *
 *              class provides an easy means of initializing and interacting  *
 *              with the IQS7221E device from an Arduino.                     *
 * @author      Azoteq PTY Ltd                                                *
 * @version     v1.2                                                          *
 * @date        2023                                                          *
 * ========================================================================== *
 * @attention   Makes use of the following standard Arduino libraries:        *
 * - Arduino.h -> Included in IQS7221E.h, comes standard with Arduino         *
 * - Wire.h    -> Included in IQS7221E.h, comes standard with Arduino         *
 ******************************************************************************/

/* Include Files */
#include "IQS7221E.h"

/* Private Global Variables */
bool iqs7221e_deviceRDY = false;
uint8_t iqs7221e_ready_pin;

/* Private Functions */
void iqs7221e_ready_interrupt(void);

/******************************************************************************/
/*                                CONSTRUCTORS                                */
/******************************************************************************/
IQS7221E::IQS7221E()
{
}

/******************************************************************************/
/*                               PUBLIC METHODS                               */
/******************************************************************************/

/**
 * @name    begin
 * @brief   A method to initialize the IQS7221E device with the device address
 *          and ready pin specified by the user.
 * @param   deviceAddress -> The address of the IQS7221E device.
 * @param   readyPin      -> The Arduino pin which is connected to the ready
 *                           pin of the IQS7221E device. Should be an 
 *                           interrupt-capable pin.
 * @retval  None.
 * @note    - Receiving a true return value does not mean that initialization 
 *            was successful. 
 *          - Receiving a true return value only means that the IQS
 *            device responded to the request for communication.
 *          - Receiving a false return value means that initialization did not
 *            take place at all. If communication is successfully established 
 *            then it is unlikely that initialization will fail.
 */
void IQS7221E::begin(uint8_t deviceAddressIn, uint8_t readyPinIn)
{
  // Initialize I2C communication here, since this library can't function without it.
  Wire.begin();
  Wire.setClock(400000);

  _deviceAddress = deviceAddressIn;
  iqs7221e_ready_pin = readyPinIn;
  attachInterrupt(digitalPinToInterrupt(iqs7221e_ready_pin), iqs7221e_ready_interrupt, CHANGE);

  /* Initialize "running" and "init" state machine variables. */
  iqs7221e_state.state      = IQS7221E_STATE_START;
  iqs7221e_state.init_state = IQS7221E_INIT_VERIFY_PRODUCT;
}

/**
 * @name    init
 * @brief   A method that runs through a normal start-up routine to set up the
 *          IQS7221E with the desired settings from the IQS7221E_init.h file.
 * @details This function must be called repeatedly until it returns true.
 *          In some stages, it uses a blocking wait until the RDY pin goes low.
 * @retval  Returns true if the full start-up routine has been completed,
 *          returns false if not.
 * @note    No false return will be given, the program will thus be stuck when
 *          one of the cases is not able to finish.
 *          See serial communication to find the ERROR case
 */
bool IQS7221E::init(void)
{
    uint16_t prod_num;
    uint8_t ver_maj, ver_min;

    switch (iqs7221e_state.init_state)
    {
    /* Verifies product number to determine if the correct device is connected for 
    this example */
    case IQS7221E_INIT_VERIFY_PRODUCT:
      if(iqs7221e_deviceRDY)
      {
        Serial.println("\tIQS7221E_INIT_VERIFY_PRODUCT");
        prod_num = getProductNum(RESTART);
        ver_maj = getmajorVersion(RESTART);
        ver_min = getminorVersion(STOP);
        Serial.print("\t\tProduct number is: ");
        Serial.print(prod_num);
        Serial.print(" v");
        Serial.print(ver_maj);
        Serial.print(".");
        Serial.println(ver_min);
        if(prod_num == IQS7221E_PRODUCT_NUM)
        {
          Serial.println("\t\tIQS7221E Release UI Confirmed!");
          iqs7221e_state.init_state = IQS7221E_INIT_READ_RESET;
        }
        else
        {
          Serial.println("\t\tDevice is not a IQS7221E!");
          iqs7221e_state.init_state = IQS7221E_INIT_NONE;
        }
      }
    break;

    /* Verify if a reset has occurred */
    case IQS7221E_INIT_READ_RESET:
      if(iqs7221e_deviceRDY)
      {
        Serial.println("\tIQS7221E_INIT_READ_RESET");
        updateInfoFlags(RESTART);
        if (checkReset())
        {
          Serial.println("\t\tReset event occurred.");
          iqs7221e_state.init_state = IQS7221E_INIT_UPDATE_SETTINGS;
        }
        else
        {
          Serial.println("\t\t No Reset Event Detected - Request SW Reset");
          iqs7221e_state.init_state = IQS7221E_INIT_CHIP_RESET;
        }
      }
    break;

    /* Perform SW Reset */
    case IQS7221E_INIT_CHIP_RESET:
      if(iqs7221e_deviceRDY)
      {
        Serial.println("\tIQS7221E_INIT_CHIP_RESET");

        //Perform SW Reset
        SW_Reset(STOP);
        Serial.println("\t\tSoftware Reset Bit Set.");
        delay(100);
        iqs7221e_state.init_state = IQS7221E_INIT_READ_RESET;
      }
    break;

    /* Write all settings to IQS7221E from .h file */
    case IQS7221E_INIT_UPDATE_SETTINGS:
      if(iqs7221e_deviceRDY)
      {
        Serial.println("\tIQS7221E_INIT_UPDATE_SETTINGS");
        writeMM(RESTART);
        iqs7221e_state.init_state = IQS7221E_INIT_ACK_RESET;
      }
    break;

    /* Acknowledge that the device went through a reset */
    case IQS7221E_INIT_ACK_RESET:
      if(iqs7221e_deviceRDY)
      {
        Serial.println("\tIQS7221E_INIT_ACK_RESET");
        acknowledgeReset(STOP);
        iqs7221e_state.init_state = IQS7221E_INIT_ATI;
      }
      break;

    /* Run the ATI algorithm to recalibrate the device with newly added settings */
    case IQS7221E_INIT_ATI:
      if(iqs7221e_deviceRDY)
      {
        Serial.println("\tIQS7221E_INIT_ATI");
        ReATI(STOP);
        iqs7221e_state.init_state = IQS7221E_INIT_WAIT_FOR_ATI;
        Serial.println("\tIQS7221E_INIT_WAIT_FOR_ATI");
      }
    break;

    /* Read the ATI Active bit to see if the rest of the program can continue */
    case IQS7221E_INIT_WAIT_FOR_ATI:
      if(iqs7221e_deviceRDY)
      {
        if(!readATIactive())
        {
          Serial.println("\t\tDONE");
          iqs7221e_state.init_state = IQS7221E_INIT_ZERO;
        }
      }
    break;

    /* Zero the encoder angle of the iqs7221e */
    case IQS7221E_INIT_ZERO:
        if(iqs7221e_deviceRDY)
        {
            zeroNow(STOP);
            Serial.println("\tIQS7221E_INIT_ZERO");
            iqs7221e_state.init_state = IQS7221E_INIT_READ_DATA;
        }
    break;

    /* Read the latest data from the iqs7221e */
    case IQS7221E_INIT_READ_DATA:
        if(iqs7221e_deviceRDY)
        {
            Serial.println("\tIQS7221E_INIT_READ_DATA");
            queueValueUpdates();
            iqs7221e_state.init_state = IQS7221E_INIT_ACTIVATE_EVENT_MODE;
        }
    break;

    /* Turn on I2C Event mode */
    case IQS7221E_INIT_ACTIVATE_EVENT_MODE:
      if(iqs7221e_deviceRDY)
      {
        Serial.println("\tIQS7221E_INIT_ACTIVATE_EVENT_MODE");
        setEventMode(STOP);
        iqs7221e_state.init_state = IQS7221E_INIT_DONE;
      }
    break;

    /* If all operations have been completed correctly, the RDY pin can be set
     * up as an interrupt to indicate when new data is available */
    case IQS7221E_INIT_DONE:
      Serial.println("\tIQS7221E_INIT_DONE");
      new_data_available = true;
      return true;
    break;

    default:
      break;
  }
  return false;
}

/**
 * @name    run
 * @brief   Run the device ready check and read the necessary bytes when the
 *          IQS7221E has pulled low the RDY line.
 *          The new_data_available flag will be set when a ready low is received
 *          from the IQS7221E.
 * @note    queueValueUpdates can be edited by the user if other data should be
 *          read every time a RDY window is received.
 */
void IQS7221E::run(void)
{
switch (iqs7221e_state.state)
  {
    /* After a hardware reset, this is the starting position of the main running  
    state machine */
    case IQS7221E_STATE_START:
      Serial.println("IQS7221E Initialization:");
      iqs7221e_state.state = IQS7221E_STATE_INIT;
    break;

    /* Perform the initialization routine on the IQS7221E */
    case IQS7221E_STATE_INIT:
      if(init())
      {
        Serial.println("IQS7221E Initialization complete!\n");
        iqs7221e_state.state = IQS7221E_STATE_RUN;
      }
    break;

    /* Send an I2C software reset in the next RDY window */
    case IQS7221E_STATE_SW_RESET:
      if(iqs7221e_deviceRDY)
      {
        SW_Reset(STOP);
        iqs7221e_state.state = IQS7221E_STATE_RUN;
      }
    break;

    /* Continuous reset monitoring state, ensure no reset event has occurred 
    for data to be valid */
    case IQS7221E_STATE_CHECK_RESET:
      if(checkReset())
      {
        Serial.println("Reset Occurred!\n");
        new_data_available = false;
        iqs7221e_state.state = IQS7221E_STATE_START;
        iqs7221e_state.init_state = IQS7221E_INIT_VERIFY_PRODUCT;
      }
      /* A reset did not occur, move to the run state and wait for a new ready 
      window */
      else
      {
        new_data_available = true; /* No reset, thus data is valid */
        iqs7221e_state.state = IQS7221E_STATE_RUN;
      }
    break;

    /* If a RDY Window is open, read the latest values from the IQS7221E */
    case IQS7221E_STATE_RUN:
      if(iqs7221e_deviceRDY)
      {
        queueValueUpdates();
        iqs7221e_deviceRDY = false;
        new_data_available = false;
        iqs7221e_state.state = IQS7221E_STATE_CHECK_RESET;
      }
    break;
  }
}

/**
 * @name    iqs7221e_ready_interrupt
 * @brief   A method used as an interrupt function. Only activated when a falling-
 *          edge interrupt is seen on the correct Arduino interrupt pin.
 * @note    Keep this function as simple as possible to prevent stuck states and
 *          slow operations.
 */
void iqs7221e_ready_interrupt(void)
{
  if(digitalRead(iqs7221e_ready_pin))
  {
      iqs7221e_deviceRDY = false;
  }
  else
  {
      iqs7221e_deviceRDY = true;
  }
}

/**
 * @name   clearRDY
 * @brief  A method used to clear the ready interrupt bit.
 * @param  None.
 * @retval None.
 */
void IQS7221E::clearRDY(void)
{
    iqs7221e_deviceRDY = false;
}

/**
  * @name   getRDYStatus
  * @brief  A method used to retrieve the device RDY status.
  * @param  None.
  * @retval Returns the boolean IQS323 RDY state.
  *         - True when RDY line is LOW
  *         - False when RDY line is HIGH
  */
bool IQS7221E::getRDYStatus(void)
{
  return iqs7221e_deviceRDY;
}

/**
 * @name    queueValueUpdates
 * @brief   All I2C read operations in the queueValueUpdates method will be
 *          performed each time the IQS7221E opens a RDY window.
 * @note    Any Address in the memory map can be read from here.
 */
void IQS7221E::queueValueUpdates(void)
{
  uint8_t transferBytes[10]; // The array which will hold the bytes to be transferred.

  /* Read the information necessary for this specific implementation from the 
  IQS7221E */
  readRandomBytes(IQS7221E_MM_SYSTEM_FLAGS, 10, transferBytes, RESTART);

  /* Assign the system status flags to the local SYSTEM_STATUS register */
  IQSMemoryMap.SYSTEM_STATUS[0] = transferBytes[0];
  IQSMemoryMap.SYSTEM_STATUS[1] = transferBytes[1];

  /* Assign the events flags to the local EVENTS register */
  IQSMemoryMap.EVENTS[0] = transferBytes[2];
  IQSMemoryMap.EVENTS[1] = transferBytes[3];

  /* Assign the HALL UI Information to the local HALL_UI register */
  IQSMemoryMap.HALL_UI[0] = transferBytes[4];
  IQSMemoryMap.HALL_UI[1] = transferBytes[5];

  /* Assign the touch states to the local TOUCH_EVENT_STATES register */
  IQSMemoryMap.TOUCH_EVENT_STATES[0] = transferBytes[6];
  IQSMemoryMap.TOUCH_EVENT_STATES[1] = transferBytes[7];

  /* Assign the free-wheel UI information to the local FREE_WHEEL_UI register */
  IQSMemoryMap.FREE_WHEEL_UI[0] = transferBytes[8];
  IQSMemoryMap.FREE_WHEEL_UI[1] = transferBytes[9];

  /* Read the hall wheel interval. */
  readRandomBytes(IQS7221E_MM_HALL_INTERVAL, 2, transferBytes, STOP);

  /* Assign the hall UI interval into the local HALL_UI_INTERVAL register. */
  IQSMemoryMap.HALL_UI_INTERVAL[0] = transferBytes[0];
  IQSMemoryMap.HALL_UI_INTERVAL[1] = transferBytes[1];
}

/**
  * @name	  readATIactive
  * @brief  A method that checks if the ATI routine is still active
  * @param  None.
  * @retval Returns true if the ATI_ACTIVE_BIT is cleared, false if the 
  *         ATI_ACTIVE_BIT is set.
  * @note   If the ATI routine is active the channel states (NONE, PROX, TOUCH) 
  *         might exhibit unwanted behaviour. Thus it is advised to wait for 
  *         the routine to complete before continuing.
  */
bool IQS7221E::readATIactive(void)
{
  /* Read the Info flags */
  updateInfoFlags(STOP);

  /* Return the ATI Active status */
  return getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7221E_ATI_ACTIVE_BIT);
}

/**
 * @name	  checkReset
 * @brief   A method that checks if the device has reset and returns the reset
 *          status.
 * @retval  Returns true if a reset has occurred, false if no reset has
 *          occurred.
 * @note    If a reset has occurred the device settings should be reloaded using
 *          the begin function. After new device settings have been reloaded the
 *          acknowledge reset function can be used to clear the reset flag.
 */
bool IQS7221E::checkReset(void)
{
   /* Perform a bitwise AND operation inside getBIT with the DEVICE_RESET_BIT 
  to return the reset status */
  return getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7221E_DEVICE_RESET_BIT);
}

/**
 * @name	  checkProductNum
 * @brief   Read and return the device product number.
 * @param   stopOrRestart -> Specifies whether the communications window must 
 *                           be kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @retval  Returns the product number as a 16-bit unsigned integer value.
 * @note    If the product is not correctly identified an appropriate messages
 *          should be displayed.
 */
uint16_t IQS7221E::getProductNum(bool stopOrRestart)
{
    uint8_t transferBytes[2];	  // A temporary array to hold the read bytes
    uint16_t prodNumReturn;     // The 16-bit return value.

    // Read the Device info from the IQS7221E.
    readRandomBytes(IQS7221E_MM_PROD_NUM, 2, transferBytes, stopOrRestart);

    // Construct the 16-bit return value.
    prodNumReturn = (uint16_t)(transferBytes[0]);
    prodNumReturn |= (uint16_t)(transferBytes[1] << 8);

    // Return the product number
    return prodNumReturn;
}

/**
 * @name	  getmajorVersion
 * @brief   A method that reads the device firmware major version number.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @retval Returns the major version number as an 8-bit unsigned integer value.
 */
uint8_t IQS7221E::getmajorVersion(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // A temporary array to hold the read bytes

    // Read the Device info from the IQS7221E.
    readRandomBytes(IQS7221E_MM_MAJOR_VERSION_NUM, 2, transferBytes, stopOrRestart);

    // Return the major firmware version number value.
    return transferBytes[0];
}

/**
 * @name	  getminorVersion
 * @brief   A method that reads the device firmware minor version number.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
  * @retval Returns the minor version number as an 8-bit unsigned integer value.
 */
uint8_t IQS7221E::getminorVersion(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // A temporary array to hold the read bytes

    // Read the Device info from the IQS7221E.
    readRandomBytes(IQS7221E_MM_MINOR_VERSION_NUM, 2, transferBytes, stopOrRestart);

    // Return the minor firmware version number value.
    return transferBytes[0];
}

/**
 * @name	  acknowledgeReset
 * @brief   A method that clears the Show Reset flag by setting the ACK_RESET
 *          bit.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    If a reset has occurred the device settings should be reloaded using
 *          the begin function. After new device settings have been reloaded,
 *          this method should be used to clear the reset bit.
 * @note    To acknowledge reset, IQS7221E_ACK_RESET_BIT of SYSTEM SETTINGS 
 *          is set.
 */
void IQS7221E::acknowledgeReset(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // A temporary array to hold the bytes transferred.

    // Read the System Flags from the IQS7221E
    // These must be read first in order not to change any settings.
    readRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferBytes, RESTART);
    // Write the Ack Reset bit to 1 to clear the Show Reset Flag.
    transferBytes[0] = setBit(transferBytes[0], IQS7221E_ACK_RESET_BIT);
    // Write the new byte to the System Flags address.
    writeRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferBytes, stopOrRestart);
}

/**
 * @name    ReATI
 * @brief   A method that sets the REDO_ATI_BIT in order to force the IQS7221E
 *          device to run the Automatic Tuning Implementation (ATI) routine.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To perform ATI, bit 1 of SYSTEM SETTINGS is set.
 */
void IQS7221E::ReATI(bool stopOrRestart)
{
    uint8_t transferByte[2]; // Array to store the bytes transferred.

    readRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferByte, RESTART);
    // Mask the settings with the REDO_ATI_BIT.
    transferByte[0] = setBit(transferByte[0], IQS7221E_FORCE_ATI_BIT);
    // Write the new byte to the required device.
    writeRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferByte, stopOrRestart);
}

/**
 * @name    SW_Reset
 * @brief   A method that sets the SW RESET bit to force the IQS7221E
 *          device to do a SW reset.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To perform SW Reset, bit 2 in SYSTEM SETTINGS is set.
 */
void IQS7221E::SW_Reset(bool stopOrRestart)
{
    uint8_t transferByte[2]; // Array to store the bytes transferred.

    readRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferByte, RESTART);
    // Mask the settings with the SW_RESET_BIT.
    transferByte[0] = setBit(transferByte[0], IQS7221E_SW_RESET_BIT);
    // Write the new byte to the required device.
    writeRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferByte, stopOrRestart);
}

/**
 * @name    setEventMode
 * @brief   A method to set the IQS7221E device into event mode.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To enter event mode, IQS7221E_EVENT_MODE_BIT of SYSTEM SETTINGS 
 *          is set.
 */
void IQS7221E::setEventMode(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // The array which will hold the bytes which are transferred.

    // First read the bytes at the memory address so that they can be preserved.
    readRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferBytes, RESTART);
    // // Clear the STREAM_EN_BIT in SYSTEM_SETTINGS lower byte
    transferBytes[0] = setBit(transferBytes[0], IQS7221E_STREAM_EN_BIT);
    // Set the EVENT_MODE_BIT in UI_SETTINGS
    transferBytes[1] = setBit(transferBytes[1], IQS7221E_EVENT_MODE_BIT);
    // Write the bytes back to the device
    writeRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferBytes, stopOrRestart);
}

/**
 * @name    setStreamMode
 * @brief   A method to set the IQS7221E device into streaming mode.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To enter event mode, IQS7221E_EVENT_MODE_BIT of SYSTEM SETTINGS 
 *          is cleared, and IQS7221E_STREAM_EN_BIT of SYSTEM SETTINGS is set.
 */
void IQS7221E::setStreamMode(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // The array which will hold the bytes transferred.

    // First read the bytes at the memory address so that they can be preserved.
    readRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferBytes, RESTART);
    // Set the STREAM_EN_BIT in SYSTEM_SETTINGS lower byte
    transferBytes[0] = setBit(transferBytes[0], IQS7221E_STREAM_EN_BIT);
    // Clear the EVENT_MODE bit in SYSTEM_SETTINGS upper byte
    transferBytes[1] = clearBit(transferBytes[1], IQS7221E_EVENT_MODE_BIT);
    // Write the bytes back to the device
    writeRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 2, transferBytes, stopOrRestart);
}

/**
 * @name    zeroNow
 * @brief   A method to zero the encoder angle of the IQS7221E.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    To run the zero-now command, set IQS7221E_ZERO_NOW_BIT of 
 *          HALL UI SETTINGS.
 */
void IQS7221E::zeroNow(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // The array which will hold the bytes transferred.

    // First read the bytes at the memory address so that they can be preserved.
    readRandomBytes(IQS7221E_MM_HALL_UI_SETTINGS, 2, transferBytes, RESTART);
    // Set the ZERO_NOW_BIT in HALL UI SETTINGS lower byte
    transferBytes[0] = setBit(transferBytes[0], IQS7221E_ZERO_NOW_BIT);
    // Write the bytes back to the device
    writeRandomBytes(IQS7221E_MM_HALL_UI_SETTINGS, 2, transferBytes, stopOrRestart);
}

/**
 * @name    updateInfoFlags
 * @brief   A method which reads the IQS7221E info flags and assigns them to the
 *          local IQSMemoryMap.SYSTEM_STATUS register.
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    The local SYSTEM_STATUS is altered with the new value of the info flags
 *          register retrieved from the IQS7221E.
 */
void IQS7221E::updateInfoFlags(bool stopOrRestart)
{
    uint8_t transferBytes[2]; // The array which will hold the bytes to be transferred.

    // Read the info flags.
    readRandomBytes(IQS7221E_MM_SYSTEM_FLAGS, 2, transferBytes, stopOrRestart);
    // Assign the info flags to the local SYSTEM_STATUS register.
    IQSMemoryMap.SYSTEM_STATUS[0] = transferBytes[0];
    IQSMemoryMap.SYSTEM_STATUS[1] = transferBytes[1];
}

/**
 * @name    get_PowerMode
 * @brief   A method that reads the local SYSTEM_STATUS register and returns 
 *          the current power mode.
 * @retval  Returns the iqs7221e_power_modes state the device is in.
 * @note    See Datasheet on power mode options and timeouts.
 *          High Accuracy, Normal Power, Low Power, Ultra Low Power (ULP), Halt.
 */
iqs7221e_power_modes IQS7221E::get_PowerMode(void)
{
    uint8_t buffer = getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7221E_POWER_MODE_BIT_0);
    buffer += getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7221E_POWER_MODE_BIT_1) << 1;
    buffer += getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7221E_POWER_MODE_BIT_2) << 2;
    buffer += getBit(IQSMemoryMap.SYSTEM_STATUS[0], IQS7221E_POWER_MODE_BIT_3) << 3;

    if (buffer == IQS7221E_HIGH_ACCURACY_BITS)
    {
        // High-accuracy bit is set while touch debouncing is active.
        // This debouncing period is very brief and can be ignored
        if ( getBit(IQSMemoryMap.TOUCH_EVENT_STATES[1], IQS7221E_TOUCH_DEBOUNCE_BIT) != 0)
        {
            return IQS7221E_NORMAL_POWER;
        }
        return IQS7221E_HIGH_ACCURACY;
    }
    else if (buffer == IQS7221E_NORMAL_POWER_BITS)
    {
        return IQS7221E_NORMAL_POWER;
    }
    else if (buffer == IQS7221E_LOW_POWER_BITS)
    {
        return IQS7221E_LOW_POWER;
    }
    else if (buffer == IQS7221E_ULP_BITS)
    {
        return IQS7221E_ULP;
    }
    else if (buffer == IQS7221E_HALT_BITS)
    {
        return IQS7221E_HALT_MODE;
    }
    else
    {
      return IQS7221E_UNKNOWN_POWER_MODE;
    }
}

/**
 * @name    button_touchState
 * @brief   A method that checks the local TOUCH_EVENT_STATES if the button is
 *          in a touch state.
 * @retval  Returns true if a touch is active and false if there is no touch.
 * @note    The IQS7221E only have 1 touch output channel.
 */
bool IQS7221E::button_touchState(void)
{
    return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7221E_TOUCH_STATE_BIT);
}

/**
 * @name    button_proxState
 * @brief   A method that checks the local TOUCH_EVENT_STATES register if the 
 *          button is in a proximity state.
 * @retval  Returns true if proximity is active and false if there is no 
 *          proximity.
 * @note    The IQS7221E only have 1 prox output channel.
 */
bool IQS7221E::button_proxState(void)
{
    return getBit(IQSMemoryMap.TOUCH_EVENT_STATES[0], IQS7221E_PROX_STATE_BIT);
}

/**
 * @name    quad_interval_changed
 * @brief   A method that checks the Hall UI flags if there was a change in
 *          interval.
 * @retval  Returns true if an interval change occurred.
 * @note    The IQS7221E quadrature out can also be determined with the
 *          dedicated quadrature GPIOs.
 */
bool IQS7221E::quad_interval_changed(void)
{
    return getBit(IQSMemoryMap.HALL_UI[0], IQS7221E_HALL_INTERVAL_CHANGED_BIT);
}

/**
 * @name    quad_interval_direction
 * @brief   A method that reads the local HALL_UI register for the Hall UI flags
 *          and determines in what direction the quadrature moved.
 * @retval  Returns true if a positive interval change occurred.
 * @note    The IQS7221E quadrature out can also be determined with the
 *          dedicated quadrature GPIOs.
 */
bool IQS7221E::quad_interval_direction(void)
{
    return getBit(IQSMemoryMap.HALL_UI[0], IQS7221E_HALL_DIRECTION_BIT);
}

/**
 * @name    quad_interval_value
 * @brief   A method that reads and returns the current interval value.
 * @retval  Returns the 16-bit unsigned integer value of the current quadrature 
 *          interval.
 * @note    The IQS7221E quadrature out can also be determined with the
 *          dedicated quadrature GPIOs.
 */
uint16_t IQS7221E::quad_interval_value(void)
{
    return (IQSMemoryMap.HALL_UI_INTERVAL[0] + (IQSMemoryMap.HALL_UI_INTERVAL[1] << 8));
}

/******************************************************************************/
/*                           ADVANCED PUBLIC METHODS                          */
/******************************************************************************/

/**
 * @name    writeMM
 * @brief   Function to write the whole memory map to the device (writable)
 *          registers
 * @param   stopOrRestart -> Specifies whether the communications window must be
 *                           kept open or must be closed after this action.
 *                           Use the STOP and RESTART definitions.
 * @note    Requires the `IQS7221E_init.h` file with the defined settings.
 */
void IQS7221E::writeMM(bool stopOrRestart)
{
    // Temporary array which holds the bytes to be transferred.
    uint8_t transferBytes[32];

    /* Change the Global Device Settings */
    /* Memory Map Position 0x60 - 0x6F */
    transferBytes[0] = UI_SETTINGS;
    transferBytes[1] = UI_SETTINGS_1;
    transferBytes[2] = I2C_SETTINGS;
    transferBytes[3] = OPEN_60_3;
    transferBytes[4] = HA_REPORT_0;
    transferBytes[5] = HA_REPORT_1;
    transferBytes[6] = NP_REPORT_0;
    transferBytes[7] = NP_REPORT_1;
    transferBytes[8] = LP_REPORT_0;
    transferBytes[9] = LP_REPORT_1;
    transferBytes[10] = ULP_REPORT_0;
    transferBytes[11] = ULP_REPORT_1;
    transferBytes[12] = ULP_TIMEOUT_0;
    transferBytes[13] = ULP_TIMEOUT_1;
    transferBytes[14] = EVENTMASK;
    transferBytes[15] = OPEN_0;
    transferBytes[16] = QUADRATURE_DELAY_0;
    transferBytes[17] = QUADRATURE_DELAY_1;
    transferBytes[18] = QUADRATURE_MODE_0;
    transferBytes[19] = QUADRATURE_MODE_1;
    transferBytes[20] = WDT_TIMEOUT_0;
    transferBytes[21] = WDT_TIMEOUT_1;
    transferBytes[22] = PS_DELAY_0;
    transferBytes[23] = PS_DELAY_1;
    transferBytes[24] = IR_ADD_0;
    transferBytes[25] = IR_ADD_1;
    transferBytes[26] = IR_DATA_0;
    transferBytes[27] = IR_DATA_1;
    transferBytes[28] = IW_ADD_0;
    transferBytes[29] = IW_ADD_1;
    transferBytes[30] = IW_DATA_0;
    transferBytes[31] = IW_DATA_1;
    writeRandomBytes(IQS7221E_MM_SYSTEM_SETTINGS, 32, transferBytes, RESTART);
    Serial.println("\t\t1. Write System Settings");

    /* Change the Hall UI Settings */
    /* Memory Map Position 0x70 - 0x7A */
    transferBytes[0] = WHEEL_TO_MAGNET_ANGLE_0;
    transferBytes[1] = WHEEL_TO_MAGNET_ANGLE_1;
    transferBytes[2] = WHEEL_TO_MAGNET_ANGLE_2;
    transferBytes[3] = WHEEL_TO_MAGNET_ANGLE_3;
    transferBytes[4] = FILTER_SWITCH_DELAY_0;
    transferBytes[5] = FILTER_SWITCH_DELAY_1;
    transferBytes[6] = OPEN_1;
    transferBytes[7] = OPEN_2;
    transferBytes[8] = INTERVAL_HYSTERESIS_0;
    transferBytes[9] = INTERVAL_HYSTERESIS_1;
    transferBytes[10] = NUMBER_INTERVALS_0;
    transferBytes[11] = NUMBER_INTERVALS_1;
    transferBytes[12] = HIGH_ACCURACY_TIMEOUT_0;
    transferBytes[13] = HIGH_ACCURACY_TIMEOUT_1;
    transferBytes[14] = STATIONARY_TIMEOUT_0;
    transferBytes[15] = STATIONARY_TIMEOUT_1;
    transferBytes[16] = SLOW_FILTER_BETA;
    transferBytes[17] = FAST_FILTER_BETA;
    transferBytes[18] = LP_FILTER_BETA;
    transferBytes[19] = AUTO_ZERO_BETA;
    transferBytes[20] = GENERAL_HALL_SETTINGS_0;
    transferBytes[21] = GENERAL_HALL_SETTINGS_1;
    writeRandomBytes(IQS7221E_MM_WHEEL_T_MAGNET_ANGLE, 22, transferBytes, RESTART);
    Serial.println("\t\t2. Write Hall UI Settings");

    /* Change the Hall Plate Settings */
    /* Memory Map Position 0x80 - 0x86 */
    transferBytes[0] = PROX_A_ATI_0;
    transferBytes[1] = PROX_A_ATI_1;
    transferBytes[2] = PROX_B_ATI_0;
    transferBytes[3] = PROX_B_ATI_1;
    transferBytes[4] = A0_HALLPLATE_BIAS;
    transferBytes[5] = B1_HALLPLATE_BIAS;
    transferBytes[6] = A1_HALLPLATE_BIAS;
    transferBytes[7] = B0_HALLPLATE_BIAS;
    transferBytes[8] = A0_LINEAR_OFFSET;
    transferBytes[9] = B1_LINEAR_OFFSET;
    transferBytes[10] = A1_LINEAR_OFFSET;
    transferBytes[11] = B0_LINEAR_OFFSET;
    transferBytes[12] = CTRL0_SETTINGS_0;
    transferBytes[13] = CTRL0_SETTINGS_1;
    transferBytes[14] = CTRL1_SETTINGS_0;
    writeRandomBytes(IQS7221E_MM_HP_FINE_COURSE_MULTI, 15, transferBytes, RESTART);
    Serial.println("\t\t3. Write Hall Plate Settings");

    /* Change the Freewheel Settings */
    /* Memory Map Position 0x90 - 0x98 */
    transferBytes[0] = FILTER_BETAS_0;
    transferBytes[1] = FILTER_BETAS_1;
    transferBytes[2] = FREEWHEEL_FRICTION_0;
    transferBytes[3] = FREEWHEEL_FRICTION_1;
    transferBytes[4] = FREEWHEEL_DAMPING_0;
    transferBytes[5] = FREEWHEEL_DAMPING_1;
    transferBytes[6] = FREEWHEEL_INERTIA_0;
    transferBytes[7] = FREEWHEEL_INERTIA_1;
    transferBytes[8] = FREEWHEEL_MIN_START_0;
    transferBytes[9] = FREEWHEEL_MIN_START_1;
    transferBytes[10] = FREEWHEEL_STOP_0;
    transferBytes[11] = FREEWHEEL_STOP_1;
    transferBytes[12] = FREEWHEEL_FILTERSPEED_0_0;
    transferBytes[13] = FREEWHEEL_FILTERSPEED_0_1;
    transferBytes[14] = FREEWHEEL_FILTERSPEED_1_0;
    transferBytes[15] = FREEWHEEL_FILTERSPEED_1_1;
    transferBytes[16] = STOP_THR;
    transferBytes[17] = FORWARD_THR;
    transferBytes[18] = REVERSE_THR;
    transferBytes[19] = OPEN_90_19;
    transferBytes[20] = MOVEMENT_THRESHOLD_STOP_0;
    transferBytes[21] = MOVEMENT_THRESHOLD_STOP_1;
    transferBytes[22] = STATIONARY_DETECT_SPEED_0;
    transferBytes[23] = STATIONARY_DETECT_SPEED_1;
    writeRandomBytes(IQS7221E_MM_FREEWHEEL_FILTER_BETA, 24, transferBytes, RESTART);
    Serial.println("\t\t4. Write Freewheel Settings");

    /* Change the Hall ATI Settings */
    /* Memory Map Position 0xA0 - 0xA2 */
    transferBytes[0] = MINIMUM_TARGET_0;
    transferBytes[1] = MINIMUM_TARGET_1;
    transferBytes[2] = MAXIMUM_TARGET_0;
    transferBytes[3] = MAXIMUM_TARGET_1;
    transferBytes[4] = ATI_FRAC;
    transferBytes[5] = RE_BLOCK;
    writeRandomBytes(IQS7221E_MM_TARGET_MIN, 6, transferBytes, RESTART);
    Serial.println("\t\t5. Write Hall ATI Settings");

    /* Change the Button Settings */
    /* Memory Map Position 0xB0 - 0xB6 */
    transferBytes[0] = BUTTON_PROX_TIMEOUT_0;
    transferBytes[1] = BUTTON_PROX_TIMEOUT_1;
    transferBytes[2] = BUTTON_TOUCH_TIMEOUT_0;
    transferBytes[3] = BUTTON_TOUCH_TIMEOUT_1;
    transferBytes[4] = BUTTON_DORMANCY_TIMEOUT_0;
    transferBytes[5] = BUTTON_DORMANCY_TIMEOUT_1;
    transferBytes[6] = ATI_TIMEOUT_0;
    transferBytes[7] = ATI_TIMEOUT_1;
    transferBytes[8] = FILTER_COUNT_BETA_NP;
    transferBytes[9] = FILTER_COUNT_BETA_LP;
    transferBytes[10] = FILTER_LTA_BETA_NP;
    transferBytes[11] = FILTER_LTA_BETA_LP;
    transferBytes[12] = FILTER_LTA_BETA_FAST_NP;
    transferBytes[13] = FILTER_LTA_BETA_FAST_LP;
    transferBytes[14] = FILTER_LTA_BETA_FAST_BAND;
    transferBytes[15] = RESEED_ATI_BAND;
    writeRandomBytes(IQS7221E_MM_BUTTON_PROX_TIMEOUT, 16, transferBytes, RESTART);
    Serial.println("\t\t6. Write Button Settings");

    /* Change the Button Sensor Settings */
    /* Memory Map Position 0xC0 - 0xC7 */
    transferBytes[0] = GENERAL_SENSOR_SETTINGS;
    transferBytes[1] = SENSOR_CONV_FREQ_FRAC;
    transferBytes[2] = SENSOR_CONV_FREQ_PERIOD;
    transferBytes[3] = GENERAL_SENSOR_SETTINGS2;
    transferBytes[4] = SENSOR_TX_SEL0;
    transferBytes[5] = SENSOR_TX_SEL1;
    transferBytes[6] = SENSOR_GENERALS;
    transferBytes[7] = SENSOR_CTRL0;
    transferBytes[8] = SENSOR_ATI_BASE0;
    transferBytes[9] = SENSOR_ATI_BASE1;
    transferBytes[10] = SENSOR_ATI_TARGET0;
    transferBytes[11] = SENSOR_ATI_TARGET1;
    transferBytes[12] = SENSOR_COARSE;
    transferBytes[13] = SENSOR_FINE;
    transferBytes[14] = SENSOR_COMPENSATION;
    transferBytes[15] = SENSOR_COMPENSATION_DIVIDER;
    writeRandomBytes(IQS7221E_MM_BUTTON_SENSOR_CONTROL_0, 16, transferBytes, RESTART);
    Serial.println("\t\t7. Write Button Sensor Settings");

    /* Change the Button Threshold Settings */
    /* Memory Map Position 0xD0 - 0xD1 */
    transferBytes[0] = PROX_THRESHOLD;
    transferBytes[1] = PROX_DEBOUNCE;
    transferBytes[2] = TOUCH_THRESHOLD;
    transferBytes[3] = TOUCH_HYSTERESIS;
    writeRandomBytes(IQS7221E_MM_BUTTON_PROX_THRESHOLD, 4, transferBytes, RESTART);
    Serial.println("\t\t8. Write Button Threshold Settings");

    /* Change the I2C Settings */
    /* Memory Map Position 0xE0 - 0xE0 */
    transferBytes[0] = I2CRDYTO_0;
    transferBytes[1] = I2CRDYTO_1;
    writeRandomBytes(IQS7221E_MM_I2C_TRANS_TIMEOUT, 2, transferBytes, stopOrRestart);
    Serial.println("\t\t9. Write I2C Settings");
}

/******************************************************************************/
/*                               PRIVATE METHODS                              */
/******************************************************************************/

/**
 * @name    readRandomBytes
 * @brief   A method that reads a specified number of bytes from a specified
 *          address and saves it into a user-supplied array.
 *          This method is used by all other methods in this class which read
 *          data from the IQS7221E device.
 * @param   memoryAddress -> The memory address at which to start reading bytes
 *                           from.  See the "iqs7221e_addresses.h" file.
 * @param   numBytes      -> The number of bytes that must be read.
 * @param   bytesArray    -> The array which will store the bytes to be read,
 *                           this array will be overwritten.
 * @param   stopOrRestart -> A boolean that specifies whether the communication
 *                           window should remain open or be closed after transfer.
 *                           False keeps it open, true closes it. Use the STOP
 *                           and RESTART definitions.
 * @retval  No value is returned, however, the user-supplied array is overwritten.
 * @note    Uses standard Arduino "Wire" library which is for I2C communication.
 *          Take note that C++ cannot return an array, therefore, the array
 *          which is passed as an argument is overwritten with the required values.
 *          Pass an array to the method by using only its name, e.g. "bytesArray",
 *          without the brackets, this passes a pointer to the array.
 */
void IQS7221E::readRandomBytes(uint8_t memoryAddress,
                               uint8_t numBytes,
                               uint8_t bytesArray[],
                               bool stopOrRestart)
{
    // A simple counter to assist with loading bytes into the user-supplied array.
    uint8_t i = 0;

    // Select the device with the address of "_deviceAddress" and start communication.
    Wire.beginTransmission(_deviceAddress);
    // Send a bit asking for the "memoryAddress" register.
    Wire.write(memoryAddress);
    // Complete the selection and communication initialization.
    Wire.endTransmission(RESTART); // Restart transmission for reading that follows.
    /* The required device has now been selected and it has been told which
    register to send information from. */

    // Request "numBytes" bytes from the device which has the address: "_deviceAddress"
    do
    {
        Wire.requestFrom((int)_deviceAddress, (int)numBytes, (int)stopOrRestart);
    } while (Wire.available() == 0); // Wait for response, this sometimes takes a few attempts

    // Load the received bytes into the array until there are no more
    while (Wire.available())
    {
        // Load the received bytes into the user-supplied array
        bytesArray[i] = Wire.read();
        i++;
    }

    /* Always manually close the RDY window after a STOP is sent to prevent 
    writing while the RDY window closes */
    if(stopOrRestart == STOP)
    {
        iqs7221e_deviceRDY = false;
    }
}

/**
 * @name    writeRandomBytes
 * @brief   A method that writes a specified number of bytes to a specified
 *          address, the bytes to write are supplied using an array pointer.
 *          This method is used by all other methods of this class which
 *          write data to the IQS7221E device.
 * @param   memoryAddress -> The memory address at which to start writing the
 *                           bytes to. See the "iqs7221e_addresses.h" file.
 * @param   numBytes      -> The number of bytes that must be written.
 * @param   bytesArray    -> The array which stores the bytes which will be
 *                           written to the memory location.
 * @param   stopOrRestart -> A boolean that specifies whether the communication
 *                           window should remain open or be closed of transfer.
 *                           False keeps it open, true closes it. Use the STOP
 *                           and RESTART definitions.
 * @retval  No value is returned, only the IQS device registers are altered.
 * @note    Uses standard Arduino "Wire" library which is for I2C communication.
 *          Take note that a full array cannot be passed to a function in C++.
 *          Pass an array to the function by using only its name, e.g. "bytesArray",
 *          without the square brackets, this passes a pointer to the array.  
 *          The values to be written must be loaded into the array before
 *          passing it to the function.
 */
void IQS7221E::writeRandomBytes(uint8_t memoryAddress,
                                uint8_t numBytes,
                                uint8_t bytesArray[],
                                bool stopOrRestart)
{
    /* Select the device with the address "_deviceAddress" and start 
    communication. */
    Wire.beginTransmission(_deviceAddress);
    // Specify the memory address where the IQS7221E must start saving the data,
    // as designated by the "memoryAddress" variable.
    Wire.write(memoryAddress);
    // Write the bytes as specified in the array that "arrayAddress" pointer points.
    for (int i = 0; i < numBytes; i++)
    {
        Wire.write(bytesArray[i]);
    }
    // End the transmission, user decides to STOP or RESTART.
    Wire.endTransmission(stopOrRestart);

    /* Always manually close the RDY window after a STOP is sent to prevent 
    writing while the RDY window closes */
    if(stopOrRestart == STOP)
    {
        iqs7221e_deviceRDY = false;
    }
}

/**
  * @name   getBit
  * @brief  A method that returns the chosen bit value of the provided byte.
  * @param  data       -> byte of which a given bit value needs to be calculated.
  * @param  bit_number -> a number between 0 and 7 representing the bit in question.
  * @retval The boolean value of the specific bit requested. 
  */
bool IQS7221E::getBit(uint8_t data, uint8_t bit_number)
{
  return (data & ( 1 << bit_number )) >> bit_number;
}

/**
  * @name   setBit
  * @brief  A method that returns the chosen bit value of the provided byte.
  * @param  data       -> byte of which a given bit value needs to be calculated.
  * @param  bit_number -> a number between 0 and 7 representing the bit in question.
  * @retval Returns an 8-bit unsigned integer value of the given data byte with 
  *         the requested bit set.
  */
uint8_t IQS7221E::setBit(uint8_t data, uint8_t bit_number)
{
	return (data |= 1UL << bit_number);
}

/**
  * @name   clearBit
  * @brief  A method that returns the chosen bit value of the provided byte.
  * @param  data       -> byte of which a given bit value needs to be calculated.
  * @param  bit_number -> a number between 0 and 7 representing the bit in question.
  * @retval Returns an 8-bit unsigned integer value of the given data byte with 
  *         the requested bit cleared.
  */
uint8_t IQS7221E::clearBit(uint8_t data, uint8_t bit_number)
{
	return (data &= ~(1UL << bit_number));
}

/**
  * @name   force_I2C_communication
  * @brief  A method that writes 0xFF to open a communication window on the 
  *         IQS7221E.
  * @note   Uses standard Arduino "Wire" library which is for I2C communication.
  */
void IQS7221E::force_I2C_communication(void)
{
  /*Ensure RDY is HIGH at the moment*/
  if (!iqs7221e_deviceRDY)
  {
      /* Select the device with the address: "DEMO_IQS7221E_ADDR" and start 
      communication. */
      Wire.beginTransmission(_deviceAddress);

      /* Write to memory address 0xFF that will prompt the IQS7221E to open a 
      communication window.*/
      Wire.write(0xFF);

      /* End the transmission, user decides to STOP or RESTART. */
      Wire.endTransmission(STOP);
  }
}
