/******************************************************************************
 *                                                                            *
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        IQS7221E.h                                                    *
 * @brief       This file contains the header information for an IQS7221E     *
 *              Arduino library. The goal of the library is to provide easy   *
 *              functionality for initializing and using the Azoteq IQS7221E  *
 *              Hall-sensing device.                                          *
 * @author      Azoteq PTY Ltd                                                *
 * @version     v1.2                                                          *
 * @date        2023                                                          *
 * ========================================================================== *
 * @attention  Makes use of the following standard Arduino libraries:         *
 * - Arduino.h   -> included in IQS7221E.h, comes standard with Arduino       *
 * - Wire.h      -> Included in IQS7221E.h, comes standard with Arduino       *
 ******************************************************************************/


#ifndef IQS7221E_h
#define IQS7221E_h

// Include Files
#include "Arduino.h"
#include "Wire.h"
#include "IQS7221E_init.h"
#include "./inc/iqs7221e_addresses.h"

// Public Global Definitions
/* For use with Wire.h library. True argument with some functions closes the 
I2C communication window. */
#define STOP true    
/* For use with Wire.h library. False argument with some functions keeps the 
I2C communication window open. */
#define RESTART false 

// Device Info
#define IQS7221E_PRODUCT_NUM                0x0503

// Info Flags Byte Bits.
#define IQS7221E_ATI_ACTIVE_BIT             0
#define IQS7221E_DEVICE_RESET_BIT           1
#define IQS7221E_POWER_MODE_BIT_0           2
#define IQS7221E_POWER_MODE_BIT_1           3
#define IQS7221E_POWER_MODE_BIT_2           4
#define IQS7221E_POWER_MODE_BIT_3           5

#define IQS7221E_HIGH_ACCURACY_BITS         0b0000 // 0x00
#define IQS7221E_NORMAL_POWER_BITS          0b0001 // 0x01
#define IQS7221E_LOW_POWER_BITS             0b0010 // 0x02
#define IQS7221E_ULP_BITS                   0b0100 // 0x04
#define IQS7221E_HALT_BITS                  0b1000 // 0x08

// System Settings Bits
#define IQS7221E_STREAM_EN_BIT              0
#define IQS7221E_FORCE_ATI_BIT              1
#define IQS7221E_SW_RESET_BIT               2
#define IQS7221E_ACK_RESET_BIT              3
#define IQS7221E_EVENT_MODE_BIT             5   // 13

// Global Event Bits
#define IQS7221E_POWER_EVENT_BIT            0
#define IQS7221E_ATI_EVENT_BIT              1
#define IQS7221E_TOUCh_EVENT_BIT            2
#define IQS7221E_PROX_EVENT_BIT             3
#define IQS7221E_HALL_EVENT_BIT             4

// Hall UI Settings bits
#define IQS7221E_ZERO_NOW_BIT               4

// Touch Flag Bits
#define IQS7221E_PROX_STATE_BIT             0
#define IQS7221E_TOUCH_STATE_BIT            1
#define IQS7221E_TOUCH_DEBOUNCE_BIT         2   // 10

// Hall UI Flag Bits
#define IQS7221E_HALL_INTERVAL_CHANGED_BIT  0
#define IQS7221E_HALL_DIRECTION_BIT         1

/* Defines and structs for IQS7221E states */
/**
 * @brief  IQS7221E Init Enumeration.
 */
typedef enum
{
    IQS7221E_INIT_NONE = (uint8_t)0x00,
    IQS7221E_INIT_VERIFY_PRODUCT,
    IQS7221E_INIT_READ_RESET,
	IQS7221E_INIT_CHIP_RESET,
	IQS7221E_INIT_UPDATE_SETTINGS,
	IQS7221E_INIT_CHECK_RESET,
	IQS7221E_INIT_ACK_RESET,
	IQS7221E_INIT_ATI,
    IQS7221E_INIT_WAIT_FOR_ATI,
    IQS7221E_INIT_ZERO,
    IQS7221E_INIT_READ_DATA,
    IQS7221E_INIT_ACTIVATE_EVENT_MODE,
    IQS7221E_INIT_DONE
} iqs7221e_init_e;

typedef enum {
    IQS7221E_STATE_NONE = (uint8_t) 0x00,
    IQS7221E_STATE_START,
    IQS7221E_STATE_INIT,
    IQS7221E_STATE_SW_RESET,
    IQS7221E_STATE_CHECK_RESET,
	IQS7221E_STATE_RUN,
} iqs7221e_state_e;

typedef enum
{
    IQS7221E_CH_NONE = (uint8_t)0x00,
    IQS7221E_CH_PROX,
    IQS7221E_CH_TOUCH,
    IQS7221E_CH_UNKNOWN,
} iqs7221e_ch_states;

typedef enum
{
    IQS7221E_HIGH_ACCURACY = (uint8_t)0x00,
    IQS7221E_NORMAL_POWER,
    IQS7221E_LOW_POWER,
    IQS7221E_ULP,
    IQS7221E_HALT_MODE,
    IQS7221E_UNKNOWN_POWER_MODE
} iqs7221e_power_modes;

#pragma pack(1)
typedef struct
{
    iqs7221e_state_e state;
    iqs7221e_init_e init_state;
} iqs7221e_s;
#pragma pack(4)

/* IQS7221E Memory map data variables, only save the data that might be used 
during program runtime */
#pragma pack(1)
typedef struct
{
	/* READ ONLY */			            //  I2C Addresses:
	uint8_t VERSION_DETAILS[6]; 	    // 	0x00 -> 0x02

	uint8_t SYSTEM_STATUS[2];           // 	0x10
	uint8_t EVENTS[2]; 		            // 	0x11
	uint8_t HALL_UI[2]; 	            // 	0x12
    uint8_t TOUCH_EVENT_STATES[2];      // 	0x13
    uint8_t FREE_WHEEL_UI[2];           // 	0x14

    uint8_t HALL_UI_INTERVAL[2];        // 	0x20

	/* READ WRITE */		            //  I2C Addresses:
	uint8_t SYSTEM_SETTINGS[2]; 	    // 	0x60
} IQS7221E_MEMORY_MAP;
#pragma pack(4)

// Class Prototype
class IQS7221E
{
public:
    // Public Constructors
    IQS7221E();

    // Public Device States
    iqs7221e_s iqs7221e_state;

    // Public Variables
    IQS7221E_MEMORY_MAP IQSMemoryMap;
    bool new_data_available;

    // Public Methods
    void begin(uint8_t deviceAddressIn, uint8_t readyPinIn);
    bool init(void);
    void run(void);
    void queueValueUpdates(void);
    bool readATIactive(void);
    uint16_t getProductNum(bool stopOrRestart);
    uint8_t getmajorVersion(bool stopOrRestart);
    uint8_t getminorVersion(bool stopOrRestart);
    void acknowledgeReset(bool stopOrRestart);
    void ReATI(bool stopOrRestart);
    void SW_Reset(bool stopOrRestart);
    void writeMM(bool stopOrRestart);
    void clearRDY(void);
    bool getRDYStatus(void);
    void zeroNow(bool stopOrRestart);

    void setStreamMode(bool stopOrRestart);
    void setEventMode(bool stopOrRestart);

    void updateInfoFlags(bool stopOrRestart);
    iqs7221e_power_modes get_PowerMode(void);
    bool checkReset(void);

    bool button_touchState(void);
    bool button_proxState(void);
    bool quad_interval_changed(void);
    bool quad_interval_direction(void);
    uint16_t quad_interval_value(void);

    void force_I2C_communication(void);

private:
    // Private Variables
    uint8_t _deviceAddress;
    uint8_t _readyPin;

    // Private Methods
    void readRandomBytes(uint8_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart);
    void writeRandomBytes(uint8_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart);
    bool getBit(uint8_t data, uint8_t bit_number);
    uint8_t setBit(uint8_t data, uint8_t bit_number);
    uint8_t clearBit(uint8_t data, uint8_t bit_number);
};
#endif // IQS7221E_h
